print("import: 'yellowbrick'")
import yellowbrick

print("import: 'yellowbrick.classifier'")
import yellowbrick.classifier

print("import: 'yellowbrick.cluster'")
import yellowbrick.cluster

print("import: 'yellowbrick.contrib'")
import yellowbrick.contrib

print("import: 'yellowbrick.contrib.classifier'")
import yellowbrick.contrib.classifier

print("import: 'yellowbrick.contrib.missing'")
import yellowbrick.contrib.missing

print("import: 'yellowbrick.contrib.statsmodels'")
import yellowbrick.contrib.statsmodels

print("import: 'yellowbrick.datasets'")
import yellowbrick.datasets

print("import: 'yellowbrick.features'")
import yellowbrick.features

print("import: 'yellowbrick.gridsearch'")
import yellowbrick.gridsearch

print("import: 'yellowbrick.model_selection'")
import yellowbrick.model_selection

print("import: 'yellowbrick.regressor'")
import yellowbrick.regressor

print("import: 'yellowbrick.style'")
import yellowbrick.style

print("import: 'yellowbrick.target'")
import yellowbrick.target

print("import: 'yellowbrick.text'")
import yellowbrick.text

print("import: 'yellowbrick.utils'")
import yellowbrick.utils

